// frogga.cpp : Definiert den Einstiegspunkt fr die Konsolenanwendung.
//
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "s1d1565.h"
#include "binary.h"
#include "frogga_title.h"

// ----------------------------------------------------------------
// PORT DEFINITION
// ----------------------------------------------------------------

// Key port
#define KEY_PIN PINC
#define KEY_DDR DDRC
#define KEY_PORT PORTC


#define KEY_NULL	-1
#define KEY_UP		0
#define KEY_LEFT	1
#define KEY_DOWN	2
#define KEY_RIGHT	3
#define KEY_ENTER	4

// Key pressed interrupt
void i_timer(void);
void HandleKeys(void);


// Stop/restore interrupts
volatile unsigned char sreg_copy = 0; // SREG contains the cli/sei state bits
#define ENTER_CRITICAL() { sreg_copy = SREG; cli(); }
#define LEAVE_CRITICAL() { SREG = sreg_copy; }



SIGNAL(SIG_OVERFLOW2)
{
static char delay=0;
	++delay;
	if(delay > 30)
	{
		delay=0;
		ENTER_CRITICAL();
		i_timer();
		LEAVE_CRITICAL();
	}
}

// Timer, triggered every uhm...
SIGNAL(SIG_OVERFLOW0)
{
static char delay=0;
	++delay;
	if(delay > 10)
	{
		ENTER_CRITICAL();
		HandleKeys();
		LEAVE_CRITICAL();
	}
}

// ----------------------------------------------------------------
// KEYS
// ----------------------------------------------------------------
volatile char KEY = KEY_NULL;
volatile char KEY_VALID=0;

void HandleKeys(void)
{
	if(KEY_VALID) return;

	for (char i=0;i<5;i++)
	{
		if(i!= KEY && !(KEY_PIN & (1<<i)) )
		{
			KEY_VALID = 1;
			KEY = i;
		}
	}
	// Key has been released
	if(!KEY_VALID && KEY!=KEY_NULL && (KEY_PIN & (1<<KEY)) )
		KEY=KEY_NULL;
}

// Is Key up/down now?
char GetKeyState(char key)
{
	return !(KEY_PIN & (1<<key) );
}

// whick key has changed to keydown
char GetKey(void)
{
char k;
    if (KEY_VALID)
    {
        k = KEY;
        KEY_VALID = 0;
    }
    else
        k = KEY_NULL;
    return k;
}


// ---------------------------------
// GLOBALS
// ---------------------------------
volatile long g_rnd;        // pseudo random number (just need 1/0)
volatile uint32_t g_road[8];    // the road-lines
volatile char g_frogpos[2];// where the frogger is
volatile char g_life;      // frogs left

// ---------------------------------
// Position a gfx block somewhere
// ---------------------------------
void Sprite(char index)
{
	// graphics
	// this arrow -> points to the top (rotated by 90 acc later)
	const char all[10][8] = {
	   {0,0,0,0,0,0,0,0},
	// car >>
	   {b01101011 ,
		b01000001 ,
		b00100010 ,
		b00100010 ,
		b01000001 ,
		b01100001 ,
		b01001001 ,
		b01101011 },
	// frog
	   {b01000000 ,
		b10110100 ,
		b11111110 ,
		b01111101 ,
		b01111110 ,
		b11110101 ,
		b10110110 ,
		b01000100 },
	// wall
	   {b11000011 ,
		b11011011 ,
		b11011011 ,
		b11011011 ,
		b11011011 ,
		b00011000 ,
		b11011011 ,
		b11011011 },
	// water >>
	   {b00000000 ,
		b00000000 ,
		b11001100 ,
		b00110011 ,
		b00000000 ,
		b00000000 ,
		b00100100 ,
		b00011000 },
	// log
	   {b00111100 ,
		b01000010 ,
		b01000010 ,
		b01100010 ,
		b01111110 ,
		b01000010 ,
		b00111100 ,
		b00000000 },
	// Goal
	   {b11111111 ,
		b10000111 ,
		b00110011 ,
		b10001011 ,
		b10000011 ,
		b00000011 ,
		b10000111 ,
		b11111111 },
	// bardwalk
	   {b00010001 ,
		b11110001 ,
		b00010001 ,
		b00011111 ,
		b00010001 ,
		b11110001 ,
		b00010001 ,
		b00011111 },
	// car <<
	   {b01101011 ,
		b01001001 ,
		b01000001 ,
		b00100010 ,
		b00100010 ,
		b01100001 ,
		b01000001 ,
		b01101011 },
	// water <<
	   {b00011000 ,
		b00100100 ,
		b00000000 ,
		b00000011 ,
		b00110011 ,
		b11001100 ,
		b00000000 ,
		b00000000 }
		};
//	lcd_writebyte(pgm_read_byte(&all[(int)index][0]));
	lcd_writebyte(all[(int)index][0]);
	lcd_writebyte(all[(int)index][1]);
	lcd_writebyte(all[(int)index][2]);
	lcd_writebyte(all[(int)index][3]);
	lcd_writebyte(all[(int)index][4]);
	lcd_writebyte(all[(int)index][5]);
	lcd_writebyte(all[(int)index][6]);
	lcd_writebyte(all[(int)index][7]);
}

// ---------------------------------
// Show a title screen
// ---------------------------------
void lcd_show_image(const char* pImg_p, char x_offset)
{
	char temp;
	for (char y=0;y<8;y++)
	{
		lcd_setaddress(x_offset, y);
		for (register unsigned char x=0; x<128-x_offset; x++)
		{
			memcpy_P(&temp, &frogga_title[(y<<7)+x], 1);
			lcd_writebyte(temp);
		}
		
	}
}




#define ROAD_AT(x,y) ((g_road[(int)y] >> x) & 0x01)

volatile char g_Update = 0;
volatile char g_road_shifting=0; // how may pixels the road rows have to be shifted

// ---------------------------------
// UpdateDisplay - show it all
// ---------------------------------
void UpdateDisplay()
{
char y,x, id, leftright;
static char dead_at_y[8] = {3, 4,9, 1, 8,1,8, 1};
static char safe_at_y[8] = {6, 5,5, 7, 0,0,0, 7};

	if(g_Update) return;
	g_Update=1;
	leftright=1;

	// Display road + frog
	for(y=0; y<8; ++y)
	{
		char start= 0;
		leftright=1-leftright;
		if(y!=0 && y!=3 && y!=7) // moving traffic
		{
			char spacer = 0;
			if(leftright)
			{
				spacer = g_road_shifting;
				start=1; // leave the left-out shifted cars
			}
			else
			{
				spacer = 8-g_road_shifting;
			}
			// start line and clear junk bits from last draw
			// also increase the x value for drawing
			lcd_setaddress(0,y);
			for(x=0; x<spacer; ++x) lcd_writebyte(0);
		}
		else // static traffic line
		{
			if(y)
				lcd_setaddress(0, y);
			else
			{
				lcd_setaddress(24, y);
				start=3;
			}
		}
		for(x=start; x<16; ++x)
		{
			if (ROAD_AT(x,y))
				id=dead_at_y[(int)y];
			else
				id=safe_at_y[(int)y];

			// if(x==g_frogpos[0] && y==g_frogpos[1]) id=2; // remove if smooth
			Sprite(id);
		}
		// inline the frog right after the line has been drawn - least flicker
		if (y == g_frogpos[1])
		{
			// Draw Frog over smooth road
			lcd_setaddress(g_frogpos[0]<<3, g_frogpos[1]);
			Sprite(2);
		}
	}
	// Life - display
	lcd_setaddress(0, 0);
	Sprite (2); // frog
	lcd_writechar('x');
	lcd_writechar(g_life + '0');

	g_Update=0;
}


// ---------------------------------
// NewGame
// ---------------------------------
void NewGame()
{
	g_life = 3;
	g_rnd = 0x1b6db7;
	g_road[0] = b11011111 | ((long)b10110110 << 8);
	g_road[1] = 0x66666666; // 1010101010...
	g_road[2] = 0x9249249; // 100100100...
	g_road[3] = 0;
	g_road[4] = 19599599;   // 1100101001100...
	g_road[5] = 0x33333333;		// 11001100
	g_road[6] = 0x22222222;     // 10001000...
	g_road[7] = 0;
	g_frogpos[0] = 8;
	g_frogpos[1] = 7;

	// Titlescreen
	lcd_clear();
	lcd_show_image(frogga_title, 0);

	// Wait for ENTER
	while(GetKey() != KEY_ENTER)
	{
		if(g_road_shifting==0)
			lcd_show_image(frogga_title, 0);
		else if(g_road_shifting == 4)
		{
			lcd_setaddress(16, 7);
			lcd_string_P ("PRESS START!");
		}
	}
	
	// scroll out titlescreen
	for(unsigned char x=0; x<128; ++x)
	{
		lcd_show_image(frogga_title, x);

		//Clear left column:
		for (char y=0;y<8;y++)
		{
			lcd_setaddress(x,y);
			lcd_writebyte(0);
		};
		asm volatile("nop"::);		
	}
	lcd_clear();
}


// ---------------------------------
// CheckCollision - and reset frog/new game
// ---------------------------------
void CheckCollision()
{
	if (ROAD_AT(g_frogpos[0], g_frogpos[1]))
	{
		// Splat
		g_frogpos[0] = 10;
		g_frogpos[1] = 7;
		--g_life;
		if (g_life<0) NewGame();
	}
}

// ---------------------------------
// timer interrupt - triggered every 2 seonds or so...
// ---------------------------------
void i_timer()
{
char y, i;
uint32_t b;
	g_road_shifting++;
	if(g_road_shifting>7)
	{
		g_road_shifting=0;

		// shift every car-row left/right and insert
		// the bit that has been shifted out
		i=0;
		++g_rnd;
		// rows 1-6 (0 and 7 stay fixed)
		for(y=1; y<7; ++y)
		{
			i=1-i;
			if(i) // left moving
			{
				b=(g_road[(int)y]>>30) & 1;
				g_road[(int)y]<<=1;
				g_road[(int)y]|= b;
			}
			else // right moving
			{
				b=g_road[(int)y] & 1;
				g_road[(int)y]>>=1;
				g_road[(int)y]|= (b<<30);
			}
		}
		CheckCollision();
	}
}


int main(int argc, char* argv[])
{
char dir=1;
    lcd_init();
	KEY_DDR = 0x00; KEY_PORT = 0xff; // input, pull-up


	//Enable 8-Bit Counter0 overflow Interrupt
	TCCR0=(1<<CS01);
	TIMSK|=(1<<TOIE0);
	
	//Enable 8-Bit Counter2
	TCCR2=(1<<CS22)|(1<<CS21);
	TIMSK|=(1<<TOIE2);
	
	sei();

	NewGame();
	for(;;)
	{
		++g_rnd;
		// Get last pressed key
		dir = GetKey();
		if(dir!=KEY_NULL)
		{
			// Move frog
			switch(dir)
			{
			case KEY_UP:     if(g_frogpos[1]>0)  --g_frogpos[1]; break;
			case KEY_DOWN:   if(g_frogpos[1]<7)  ++g_frogpos[1]; break;
			case KEY_LEFT:   if(g_frogpos[0]>0)  --g_frogpos[0]; break;
			case KEY_RIGHT:  if(g_frogpos[0]<20) ++g_frogpos[0]; break;
			}
			// Handle frog + update display
			CheckCollision();
		}
		UpdateDisplay();
	}
	return 0;
}
